import typing
import os
import pathlib
import urllib
import md2gemini
from .static import StaticDirectoryApplication

from .base import (
    Request,
    Response,
    Status,
)


class StaticMarkdownDirectoryApplication(StaticDirectoryApplication):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def serve_static_file(self, request: Request) -> Response:
        """
        Convert a URL into a filesystem path, and attempt to serve the file
        or directory that is represented at that path.
        """
        url_path = pathlib.Path(request.path.strip("/"))

        filename = pathlib.Path(os.path.normpath(str(url_path)))
        if filename.is_absolute() or str(filename).startswith(".."):
            # Guard against breaking out of the directory
            return Response(Status.NOT_FOUND, "Not Found")

        if str(filename).startswith(self.cgi_directory):
            # CGI needs special treatment to account for extra-path
            # PATH_INFO component (RFC 3875 section 4.1.5)

            # Identify the shortest path that is not a directory
            for i in range(2, len(filename.parts) + 1):
                # Split the path into SCRIPT_NAME and PATH_INFO
                script_name = pathlib.Path(*filename.parts[:i])
                path_info = pathlib.Path(*filename.parts[i:])

                filesystem_path = self.root / script_name
                try:
                    if not filesystem_path.is_file():
                        continue
                    elif not os.access(filesystem_path, os.R_OK):
                        continue
                    elif not os.access(filesystem_path, os.X_OK):
                        continue
                    else:
                        if str(script_name) == ".":
                            request.environ["SCRIPT_NAME"] = ""
                        else:
                            request.environ["SCRIPT_NAME"] = f"/{script_name}"

                        if str(path_info) == ".":
                            request.environ["PATH_INFO"] = ""
                        else:
                            request.environ["PATH_INFO"] = f"/{path_info}"

                        # Add back the trailing slash that was stripped off
                        if request.path.endswith("/"):
                            request.environ["PATH_INFO"] += "/"

                        return self.run_cgi_script(filesystem_path, request.environ)

                except OSError:
                    # Filename too large, etc.
                    return Response(Status.NOT_FOUND, "Not Found")

        filesystem_path = self.root / filename

        try:
            if not os.access(filesystem_path, os.R_OK):
                # File not readable
                return Response(Status.NOT_FOUND, "Not Found")
        except OSError:
            # Filename too large, etc.
            return Response(Status.NOT_FOUND, "Not Found")

        if filesystem_path.is_file():
            return self.generate_response(filesystem_path)

        elif filesystem_path.is_dir():
            if request.path and not request.path.endswith("/"):
                url_parts = urllib.parse.urlparse(request.url)
                # noinspection PyProtectedMember
                url_parts = url_parts._replace(path=request.path + "/")
                return Response(Status.REDIRECT_PERMANENT, url_parts.geturl())

            index_file = filesystem_path / self.index_file
            if index_file.exists():
                return self.generate_response(index_file)

            mimetype = self.add_extra_parameters("text/gemini")
            generator = self.list_directory(url_path, filesystem_path)
            return Response(Status.SUCCESS, mimetype, generator)

        else:
            return Response(Status.NOT_FOUND, "Not Found")

    def generate_response(self, filesystem_path: pathlib.Path) -> typing.Iterator:
        mimetype = self.guess_mimetype(filesystem_path.name)
        if mimetype == "text/markdown":
            generator = self._load_md_file(filesystem_path)
            mimetype = self.add_extra_parameters("text/gemini")
        elif mimetype == "text/gemini":
            generator = self.load_file(filesystem_path)
            mimetype = self.add_extra_parameters("text/gemini")
        else:
            generator = self.load_file(filesystem_path)
            mimetype = self.add_extra_parameters(mimetype)
        return Response(Status.SUCCESS, mimetype, generator)

    def _load_md_file(self, filesystem_path: pathlib.Path) -> typing.Iterator[bytes]:
        with filesystem_path.open("r") as fp:
            gemini = md2gemini.md2gemini(fp.read(), links="paragraph")
        gemini = gemini.encode()
        chunks = int(len(gemini) / self.CHUNK_SIZE) + 1
        for i in range(chunks):
            start = self.CHUNK_SIZE * (i)
            end = self.CHUNK_SIZE * (i + 1)
            yield gemini[start:end]

    def list_directory(
        self, url_path: pathlib.Path, filesystem_path: pathlib.Path
    ) -> typing.Iterator[bytes]:
        """
        Auto-generate a text/gemini document based on the contents of the file system.
        """
        buffer = f"Directory: /{url_path}\r\n".encode()
        if url_path.parent != url_path:
            buffer += f"=>/{url_path.parent}\t..\r\n".encode()

        for file in sorted(filesystem_path.iterdir()):
            if file.name.startswith("."):
                # Skip hidden directories/files that may contain sensitive info
                continue

            encoded_path = urllib.parse.quote(str(url_path / file.name))
            if file.is_dir():
                buffer += f"=>/{encoded_path}/\t{file.name}/\r\n".encode()
            else:
                if file.stem.isdigit():
                    tz = pytz.timezone("Europe/Brussels")
                    label = datetime.datetime.fromtimestamp(int(file.stem), tz=tz)
                else:
                    label = f"{file.name}"
                buffer += f"=>/{encoded_path}\t{label}\r\n".encode()

            if len(buffer) >= self.CHUNK_SIZE:
                data, buffer = buffer[: self.CHUNK_SIZE], buffer[self.CHUNK_SIZE :]
                yield data

        if buffer:
            yield buffer
